@@ -19,8 +19,12 @@ import errno
 import logging
 import multiprocessing
 import os
+import pickle
+import resource
+import socket
 import stat
 import subprocess
+import sys
 import tempfile
 
 import fixtures
@@ -30,6 +34,8 @@ import six
 
 from oslo_concurrency import processutils
 from oslotest import mockpatch
+
+
 PROCESS_EXECUTION_ERROR_LOGGING_TEST = """#!/bin/bash
 exit 41"""
 
@@ -41,6 +47,12 @@ echo onstdout --password='"secret"'
 echo onstderr --password='"secret"' 1>&2
 exit 38"""
 
+# This byte sequence is undecodable from most encoding
+UNDECODABLE_BYTES = b'[a\x80\xe9\xff]'
+
+TRUE_UTILITY = (sys.platform.startswith('darwin') and
+                '/usr/bin/true' or '/bin/true')
+
 
 class UtilsTest(test_base.BaseTestCase):
     # NOTE(jkoelker) Moar tests from nova need to be ported. But they
@@ -64,11 +76,11 @@ class UtilsTest(test_base.BaseTestCase):
     def test_execute_with_callback(self):
         on_execute_callback = mock.Mock()
         on_completion_callback = mock.Mock()
-        processutils.execute("/bin/true")
+        processutils.execute(TRUE_UTILITY)
         self.assertEqual(0, on_execute_callback.call_count)
         self.assertEqual(0, on_completion_callback.call_count)
 
-        processutils.execute("/bin/true", on_execute=on_execute_callback,
+        processutils.execute(TRUE_UTILITY, on_execute=on_execute_callback,
                              on_completion=on_completion_callback)
         self.assertEqual(1, on_execute_callback.call_count)
         self.assertEqual(1, on_completion_callback.call_count)
@@ -85,7 +97,7 @@ class UtilsTest(test_base.BaseTestCase):
 
         self.assertRaises(IOError,
                           processutils.execute,
-                          "/bin/true",
+                          TRUE_UTILITY,
                           on_execute=on_execute_callback,
                           on_completion=on_completion_callback)
         self.assertEqual(1, on_execute_callback.call_count)
@@ -99,13 +111,13 @@ class UtilsTest(test_base.BaseTestCase):
         def preexec_fn():
             raise processutils.InvalidArgumentError()
 
-        processutils.execute("/bin/true")
+        processutils.execute(TRUE_UTILITY)
 
         expected_exception = (processutils.InvalidArgumentError if six.PY2
                               else subprocess.SubprocessError)
         self.assertRaises(expected_exception,
                           processutils.execute,
-                          "/bin/true",
+                          TRUE_UTILITY,
                           preexec_fn=preexec_fn)
 
 
@@ -195,11 +207,12 @@ exit 1
             fp = open(tmpfilename2, 'r')
             runs = fp.read()
             fp.close()
-            self.assertNotEqual(runs.strip(), 'failure', 'stdin did not '
-                                                         'always get passed '
-                                                         'correctly')
+            self.assertNotEqual('failure', 'stdin did not '
+                                           'always get passed '
+                                           'correctly',
+                                runs.strip())
             runs = int(runs.strip())
-            self.assertEqual(runs, 10, 'Ran %d times instead of 10.' % (runs,))
+            self.assertEqual(10, runs, 'Ran %d times instead of 10.' % (runs,))
         finally:
             os.unlink(tmpfilename)
             os.unlink(tmpfilename2)
@@ -221,7 +234,7 @@ exit 1
         out, err = processutils.execute('/usr/bin/env',
                                         'sh', '-c', 'pwd',
                                         cwd=tmpdir)
-        self.assertIn(six.b(tmpdir), out)
+        self.assertIn(tmpdir, out)
 
     def test_check_exit_code_list(self):
         processutils.execute('/usr/bin/env', 'sh', '-c', 'exit 101',
@@ -312,7 +325,7 @@ grep foo
 
         self.assertEqual(attempts if attempts else 1, mock.mock.call_count)
         self.assertIn('Got an OSError', fixture.output)
-        self.assertIn('errno: 11', fixture.output)
+        self.assertIn('errno: %d' % errno.EAGAIN, fixture.output)
         self.assertIn("'/usr/bin/env false'", fixture.output)
 
     def test_logging_on_communicate_error_1(self):
@@ -349,6 +362,19 @@ grep foo
         env_vars = {'SUPER_UNIQUE_VAR': 'The answer is 42'}
 
         out, err = processutils.execute('/usr/bin/env', env_variables=env_vars)
+        self.assertIsInstance(out, str)
+        self.assertIsInstance(err, str)
+
+        self.assertIn('SUPER_UNIQUE_VAR=The answer is 42', out)
+
+    def test_binary(self):
+        env_vars = {'SUPER_UNIQUE_VAR': 'The answer is 42'}
+
+        out, err = processutils.execute('/usr/bin/env',
+                                        env_variables=env_vars,
+                                        binary=True)
+        self.assertIsInstance(out, bytes)
+        self.assertIsInstance(err, bytes)
 
         self.assertIn(b'SUPER_UNIQUE_VAR=The answer is 42', out)
 
@@ -369,13 +395,97 @@ grep foo
                                 'something')
 
         self.assertEqual(38, err.exit_code)
+        self.assertIsInstance(err.stdout, six.text_type)
+        self.assertIsInstance(err.stderr, six.text_type)
         self.assertIn('onstdout --password="***"', err.stdout)
         self.assertIn('onstderr --password="***"', err.stderr)
-        self.assertEqual(err.cmd, ' '.join([tmpfilename,
-                                            'password="***"',
-                                            'something']))
+        self.assertEqual(' '.join([tmpfilename,
+                                   'password="***"',
+                                   'something']),
+                         err.cmd)
         self.assertNotIn('secret', str(err))
 
+    def execute_undecodable_bytes(self, out_bytes, err_bytes,
+                                  exitcode=0, binary=False):
+        if six.PY3:
+            code = ';'.join(('import sys',
+                             'sys.stdout.buffer.write(%a)' % out_bytes,
+                             'sys.stdout.flush()',
+                             'sys.stderr.buffer.write(%a)' % err_bytes,
+                             'sys.stderr.flush()',
+                             'sys.exit(%s)' % exitcode))
+        else:
+            code = ';'.join(('import sys',
+                             'sys.stdout.write(%r)' % out_bytes,
+                             'sys.stdout.flush()',
+                             'sys.stderr.write(%r)' % err_bytes,
+                             'sys.stderr.flush()',
+                             'sys.exit(%s)' % exitcode))
+
+        return processutils.execute(sys.executable, '-c', code, binary=binary)
+
+    def check_undecodable_bytes(self, binary):
+        out_bytes = b'out: ' + UNDECODABLE_BYTES
+        err_bytes = b'err: ' + UNDECODABLE_BYTES
+        out, err = self.execute_undecodable_bytes(out_bytes, err_bytes,
+                                                  binary=binary)
+        if six.PY3 and not binary:
+            self.assertEqual(os.fsdecode(out_bytes), out)
+            self.assertEqual(os.fsdecode(err_bytes), err)
+        else:
+            self.assertEqual(out, out_bytes)
+            self.assertEqual(err, err_bytes)
+
+    def test_undecodable_bytes(self):
+        self.check_undecodable_bytes(False)
+
+    def test_binary_undecodable_bytes(self):
+        self.check_undecodable_bytes(True)
+
+    def check_undecodable_bytes_error(self, binary):
+        out_bytes = b'out: password="secret1" ' + UNDECODABLE_BYTES
+        err_bytes = b'err: password="secret2" ' + UNDECODABLE_BYTES
+        exc = self.assertRaises(processutils.ProcessExecutionError,
+                                self.execute_undecodable_bytes,
+                                out_bytes, err_bytes, exitcode=1,
+                                binary=binary)
+
+        out = exc.stdout
+        err = exc.stderr
+        out_bytes = b'out: password="***" ' + UNDECODABLE_BYTES
+        err_bytes = b'err: password="***" ' + UNDECODABLE_BYTES
+        if six.PY3:
+            # On Python 3, stdout and stderr attributes of
+            # ProcessExecutionError must always be Unicode
+            self.assertEqual(os.fsdecode(out_bytes), out)
+            self.assertEqual(os.fsdecode(err_bytes), err)
+        else:
+            # On Python 2, stdout and stderr attributes of
+            # ProcessExecutionError must always be bytes
+            self.assertEqual(out_bytes, out)
+            self.assertEqual(err_bytes, err)
+
+    def test_undecodable_bytes_error(self):
+        self.check_undecodable_bytes_error(False)
+
+    def test_binary_undecodable_bytes_error(self):
+        self.check_undecodable_bytes_error(True)
+
+    def test_picklable(self):
+        exc = processutils.ProcessExecutionError(
+            stdout='my stdout', stderr='my stderr',
+            exit_code=42, cmd='my cmd',
+            description='my description')
+        exc_message = str(exc)
+
+        exc = pickle.loads(pickle.dumps(exc))
+        self.assertEqual('my stdout', exc.stdout)
+        self.assertEqual('my stderr', exc.stderr)
+        self.assertEqual(42, exc.exit_code)
+        self.assertEqual('my cmd', exc.cmd)
+        self.assertEqual('my description', exc.description)
+        self.assertEqual(str(exc), exc_message)
+
 
 class ProcessExecutionErrorLoggingTest(test_base.BaseTestCase):
     def setUp(self):
@@ -481,21 +591,25 @@ class FakeSshChannel(object):
         return self.rc
 
 
-class FakeSshStream(six.StringIO):
+class FakeSshStream(six.BytesIO):
     def setup_channel(self, rc):
         self.channel = FakeSshChannel(rc)
 
 
 class FakeSshConnection(object):
-    def __init__(self, rc):
+    def __init__(self, rc, out=b'stdout', err=b'stderr'):
         self.rc = rc
+        self.out = out
+        self.err = err
 
-    def exec_command(self, cmd):
-        stdout = FakeSshStream('stdout')
+    def exec_command(self, cmd, timeout=None):
+        if timeout:
+            raise socket.timeout()
+        stdout = FakeSshStream(self.out)
         stdout.setup_channel(self.rc)
-        return (six.StringIO(),
+        return (six.BytesIO(),
                 stdout,
-                six.StringIO('stderr'))
+                six.BytesIO(self.err))
 
 
 class SshExecuteTestCase(test_base.BaseTestCase):
@@ -509,10 +623,77 @@ class SshExecuteTestCase(test_base.BaseTestCase):
                           processutils.ssh_execute,
                           None, 'ls', process_input='important')
 
+    def test_timeout_error(self):
+        self.assertRaises(socket.timeout,
+                          processutils.ssh_execute,
+                          FakeSshConnection(0), 'ls',
+                          timeout=10)
+
     def test_works(self):
-        o, e = processutils.ssh_execute(FakeSshConnection(0), 'ls')
-        self.assertEqual('stdout', o)
-        self.assertEqual('stderr', e)
+        out, err = processutils.ssh_execute(FakeSshConnection(0), 'ls')
+        self.assertEqual('stdout', out)
+        self.assertEqual('stderr', err)
+        self.assertIsInstance(out, six.text_type)
+        self.assertIsInstance(err, six.text_type)
+
+    def test_binary(self):
+        o, e = processutils.ssh_execute(FakeSshConnection(0), 'ls',
+                                        binary=True)
+        self.assertEqual(b'stdout', o)
+        self.assertEqual(b'stderr', e)
+        self.assertIsInstance(o, bytes)
+        self.assertIsInstance(e, bytes)
+
+    def check_undecodable_bytes(self, binary):
+        out_bytes = b'out: ' + UNDECODABLE_BYTES
+        err_bytes = b'err: ' + UNDECODABLE_BYTES
+        conn = FakeSshConnection(0, out=out_bytes, err=err_bytes)
+
+        out, err = processutils.ssh_execute(conn, 'ls', binary=binary)
+        if six.PY3 and not binary:
+            self.assertEqual(os.fsdecode(out_bytes), out)
+            self.assertEqual(os.fsdecode(err_bytes), err)
+        else:
+            self.assertEqual(out_bytes, out)
+            self.assertEqual(err_bytes, err)
+
+    def test_undecodable_bytes(self):
+        self.check_undecodable_bytes(False)
+
+    def test_binary_undecodable_bytes(self):
+        self.check_undecodable_bytes(True)
+
+    def check_undecodable_bytes_error(self, binary):
+        out_bytes = b'out: password="secret1" ' + UNDECODABLE_BYTES
+        err_bytes = b'err: password="secret2" ' + UNDECODABLE_BYTES
+        conn = FakeSshConnection(1, out=out_bytes, err=err_bytes)
+
+        out_bytes = b'out: password="***" ' + UNDECODABLE_BYTES
+        err_bytes = b'err: password="***" ' + UNDECODABLE_BYTES
+
+        exc = self.assertRaises(processutils.ProcessExecutionError,
+                                processutils.ssh_execute,
+                                conn, 'ls',
+                                binary=binary, check_exit_code=True)
+
+        out = exc.stdout
+        err = exc.stderr
+        if six.PY3:
+            # On Python 3, stdout and stderr attributes of
+            # ProcessExecutionError must always be Unicode
+            self.assertEqual(os.fsdecode(out_bytes), out)
+            self.assertEqual(os.fsdecode(err_bytes), err)
+        else:
+            # On Python 2, stdout and stderr attributes of
+            # ProcessExecutionError must always be bytes
+            self.assertEqual(out_bytes, out)
+            self.assertEqual(err_bytes, err)
+
+    def test_undecodable_bytes_error(self):
+        self.check_undecodable_bytes_error(False)
+
+    def test_binary_undecodable_bytes_error(self):
+        self.check_undecodable_bytes_error(True)
 
     def test_fails(self):
         self.assertRaises(processutils.ProcessExecutionError,
@@ -520,13 +701,13 @@ class SshExecuteTestCase(test_base.BaseTestCase):
 
     def _test_compromising_ssh(self, rc, check):
         fixture = self.useFixture(fixtures.FakeLogger(level=logging.DEBUG))
-        fake_stdin = six.StringIO()
+        fake_stdin = six.BytesIO()
 
         fake_stdout = mock.Mock()
         fake_stdout.channel.recv_exit_status.return_value = rc
-        fake_stdout.read.return_value = 'password="secret"'
+        fake_stdout.read.return_value = b'password="secret"'
 
-        fake_stderr = six.StringIO('password="foobar"')
+        fake_stderr = six.BytesIO(b'password="foobar"')
 
         command = 'ls --password="bar"'
 
@@ -541,9 +722,9 @@ class SshExecuteTestCase(test_base.BaseTestCase):
                                     check_exit_code=check)
 
             self.assertEqual(rc, err.exit_code)
-            self.assertEqual(err.stdout, 'password="***"')
-            self.assertEqual(err.stderr, 'password="***"')
-            self.assertEqual(err.cmd, 'ls --password="***"')
+            self.assertEqual('password="***"', err.stdout)
+            self.assertEqual('password="***"', err.stderr)
+            self.assertEqual('ls --password="***"', err.cmd)
             self.assertNotIn('secret', str(err))
             self.assertNotIn('foobar', str(err))
         else:
@@ -571,3 +752,168 @@ class SshExecuteTestCase(test_base.BaseTestCase):
 
     def test_compromising_ssh6(self):
         self._test_compromising_ssh(rc=-1, check=False)
+
+
+class PrlimitTestCase(test_base.BaseTestCase):
+    # Simply program that does nothing and returns an exit code 0.
+    # Use Python to be portable.
+    SIMPLE_PROGRAM = [sys.executable, '-c', 'pass']
+
+    def soft_limit(self, res, substract, default_limit):
+        # Create a new soft limit for a resource, lower than the current
+        # soft limit.
+        soft_limit, hard_limit = resource.getrlimit(res)
+        if soft_limit <= 0:
+            soft_limit = default_limit
+        else:
+            soft_limit -= substract
+        return soft_limit
+
+    def memory_limit(self, res):
+        # Substract 1 kB just to get a different limit. Don't substract too
+        # much to avoid memory allocation issues.
+        #
+        # Use 1 GB by default. Limit high enough to be able to load shared
+        # libraries. Limit low enough to be work on 32-bit platforms.
+        return self.soft_limit(res, 1024, 1024 ** 3)
+
+    def limit_address_space(self):
+        max_memory = self.memory_limit(resource.RLIMIT_AS)
+        return processutils.ProcessLimits(address_space=max_memory)
+
+    def test_simple(self):
+        # Simple test running a program (/bin/true) with no parameter
+        prlimit = self.limit_address_space()
+        stdout, stderr = processutils.execute(*self.SIMPLE_PROGRAM,
+                                              prlimit=prlimit)
+        self.assertEqual('', stdout.rstrip())
+        self.assertEqual(stderr.rstrip(), '')
+
+    def check_limit(self, prlimit, resource, value):
+        code = ';'.join(('import resource',
+                         'print(resource.getrlimit(resource.%s))' % resource))
+        args = [sys.executable, '-c', code]
+        stdout, stderr = processutils.execute(*args, prlimit=prlimit)
+        expected = (value, value)
+        self.assertEqual(str(expected), stdout.rstrip())
+
+    def test_address_space(self):
+        prlimit = self.limit_address_space()
+        self.check_limit(prlimit, 'RLIMIT_AS', prlimit.address_space)
+
+    def test_core_size(self):
+        size = self.soft_limit(resource.RLIMIT_CORE, 1, 1024)
+        prlimit = processutils.ProcessLimits(core_file_size=size)
+        self.check_limit(prlimit, 'RLIMIT_CORE', prlimit.core_file_size)
+
+    def test_cpu_time(self):
+        time = self.soft_limit(resource.RLIMIT_CPU, 1, 1024)
+        prlimit = processutils.ProcessLimits(cpu_time=time)
+        self.check_limit(prlimit, 'RLIMIT_CPU', prlimit.cpu_time)
+
+    def test_data_size(self):
+        max_memory = self.memory_limit(resource.RLIMIT_DATA)
+        prlimit = processutils.ProcessLimits(data_size=max_memory)
+        self.check_limit(prlimit, 'RLIMIT_DATA', max_memory)
+
+    def test_file_size(self):
+        size = self.soft_limit(resource.RLIMIT_FSIZE, 1, 1024)
+        prlimit = processutils.ProcessLimits(file_size=size)
+        self.check_limit(prlimit, 'RLIMIT_FSIZE', prlimit.file_size)
+
+    def test_memory_locked(self):
+        max_memory = self.memory_limit(resource.RLIMIT_MEMLOCK)
+        prlimit = processutils.ProcessLimits(memory_locked=max_memory)
+        self.check_limit(prlimit, 'RLIMIT_MEMLOCK', max_memory)
+
+    def test_resident_set_size(self):
+        max_memory = self.memory_limit(resource.RLIMIT_RSS)
+        prlimit = processutils.ProcessLimits(resident_set_size=max_memory)
+        self.check_limit(prlimit, 'RLIMIT_RSS', max_memory)
+
+    def test_number_files(self):
+        nfiles = self.soft_limit(resource.RLIMIT_NOFILE, 1, 1024)
+        prlimit = processutils.ProcessLimits(number_files=nfiles)
+        self.check_limit(prlimit, 'RLIMIT_NOFILE', nfiles)
+
+    def test_number_processes(self):
+        nprocs = self.soft_limit(resource.RLIMIT_NPROC, 1, 65535)
+        prlimit = processutils.ProcessLimits(number_processes=nprocs)
+        self.check_limit(prlimit, 'RLIMIT_NPROC', nprocs)
+
+    def test_stack_size(self):
+        max_memory = self.memory_limit(resource.RLIMIT_STACK)
+        prlimit = processutils.ProcessLimits(stack_size=max_memory)
+        self.check_limit(prlimit, 'RLIMIT_STACK', max_memory)
+
+    def test_unsupported_prlimit(self):
+        self.assertRaises(ValueError, processutils.ProcessLimits, xxx=33)
+
+    def test_relative_path(self):
+        prlimit = self.limit_address_space()
+        program = sys.executable
+
+        env = dict(os.environ)
+        env['PATH'] = os.path.dirname(program)
+        args = [os.path.basename(program), '-c', 'pass']
+        processutils.execute(*args, prlimit=prlimit, env_variables=env)
+
+    def test_execv_error(self):
+        prlimit = self.limit_address_space()
+        args = ['/missing_path/dont_exist/program']
+        try:
+            processutils.execute(*args, prlimit=prlimit)
+        except processutils.ProcessExecutionError as exc:
+            self.assertEqual(1, exc.exit_code)
+            self.assertEqual('', exc.stdout)
+            expected = ('%s -m oslo_concurrency.prlimit: '
+                        'failed to execute /missing_path/dont_exist/program: '
+                        % os.path.basename(sys.executable))
+            self.assertIn(expected, exc.stderr)
+        else:
+            self.fail("ProcessExecutionError not raised")
+
+    def test_setrlimit_error(self):
+        prlimit = self.limit_address_space()
+
+        # trying to set a limit higher than the current hard limit
+        # with setrlimit() should fail.
+        higher_limit = prlimit.address_space + 1024
+
+        args = [sys.executable, '-m', 'oslo_concurrency.prlimit',
+                '--as=%s' % higher_limit,
+                '--']
+        args.extend(self.SIMPLE_PROGRAM)
+        try:
+            processutils.execute(*args, prlimit=prlimit)
+        except processutils.ProcessExecutionError as exc:
+            self.assertEqual(1, exc.exit_code)
+            self.assertEqual('', exc.stdout)
+            expected = ('%s -m oslo_concurrency.prlimit: '
+                        'failed to set the AS resource limit: '
+                        % os.path.basename(sys.executable))
+            self.assertIn(expected, exc.stderr)
+        else:
+            self.fail("ProcessExecutionError not raised")
+
+    @mock.patch.object(os, 'name', 'nt')
+    @mock.patch.object(processutils.subprocess, "Popen")
+    def test_prlimit_windows(self, mock_popen):
+        # We want to ensure that process resource limits are
+        # ignored on Windows, in which case this feature is not
+        # supported. We'll just check the command passed to Popen,
+        # which is expected to be unaltered.
+        prlimit = self.limit_address_space()
+        mock_popen.return_value.communicate.return_value = None
+
+        processutils.execute(
+            *self.SIMPLE_PROGRAM,
+            prlimit=prlimit,
+            check_exit_code=False)
+
+        mock_popen.assert_called_once_with(
+            self.SIMPLE_PROGRAM,
+            stdin=mock.ANY, stdout=mock.ANY,
+            stderr=mock.ANY, close_fds=mock.ANY,
+            preexec_fn=mock.ANY, shell=mock.ANY,
+            cwd=mock.ANY, env=mock.ANY)
